package pl.edu.amu.demo.life.service;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import pl.edu.amu.demo.life.dto.PersonDTO;
import pl.edu.amu.demo.life.dto.PolicyDTO;
import pl.edu.amu.demo.life.jpa.Person;
import pl.edu.amu.demo.life.jpa.Policy;
import pl.edu.amu.demo.life.repository.PersonRepository;
import pl.edu.amu.demo.life.repository.PolicyRepository;

import java.time.LocalDate;
import java.util.List;
import java.util.stream.StreamSupport;

import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import static java.util.stream.Collectors.toList;

@Service
@Slf4j
public class PolicyService {

    private final PolicyRepository policyRepository;
    private final PersonRepository personRepository;

    @Autowired
    public PolicyService(PolicyRepository policyRepository, PersonRepository personRepository) {
        this.policyRepository = policyRepository;
        this.personRepository = personRepository;
    }

    public long countPolicies() {
        log.info("Repository class: {}", policyRepository.getClass().getCanonicalName());
        var policies = policyRepository.findAll();
        var policiesCount = StreamSupport.stream(policies.spliterator(), false).count();
        log.info("{} policies found in the repository.", policiesCount);
        return policiesCount;
    }

    // Don't do that!!!
    // NEVER, EVER YOU SHOULD QUERY FOR ALL!!!1
    // Just an example!
    public List<PolicyDTO> getAllPolicies() {
        return StreamSupport.stream(policyRepository.findAll().spliterator(), false)
                .map(PolicyDTO::fromPolicy)
                .collect(toList());
    }

    public void savePolicy(PolicyDTO policyDTO) {
        Policy policyToAdd = policyDTO.toPolicy();
        if (nonNull(policyToAdd.getId())) {
            log.error("Attempt to add policy with id {}", policyToAdd.getId());
            throw new IllegalStateException("New policies must not have an id!");
        }
        if (nonNull(policyToAdd.getInsuredPerson().getId())) {
            Person person = findPerson(policyToAdd.getInsuredPerson().getId());
            policyToAdd.setInsuredPerson(person);
        }
        policyRepository.save(policyToAdd);
    }

    private Person findPerson(Long personId) {
        return personRepository.findById(personId)
                .orElseThrow(() -> new IllegalStateException(String.format("Person id not found: %d", personId)));
    }

    public void updatePolicy(Long policyId, PolicyDTO policyDTO) {
        Policy policy = policyRepository.findById(policyId)
                .orElseThrow(() -> new IllegalStateException(String.format("Policy not found: %d", policyId)));
        Policy updated = policyDTO.toPolicy();
        if (nonNull(updated.getInsuredSum()) && !policy.getInsuredSum().equals(updated.getInsuredSum())) {
            policy.setInsuredSum(updated.getInsuredSum());
        }
        if (nonNull(updated.getEndDate()) && !policy.getEndDate().equals(updated.getEndDate())) {
            if (!updated.getEndDate().isAfter(policy.getStartDate())) {
                throw new IllegalStateException("End date must be after start date!");
            }
            policy.setEndDate(updated.getEndDate());
        }
        if (nonNull(updated.getStatus()) && !policy.getStatus().equals(updated.getStatus())) {
            policy.setStatus(updated.getStatus());
            policy.setStatusChangeDate(LocalDate.now());
        }
        policyRepository.save(policy);
    }

    public void updatePerson(PersonDTO personDTO) {
        if (isNull(personDTO.getId())) {
            throw new IllegalStateException("Unable to update person without an id.");
        }
        var person = findPerson(personDTO.getId());
        if (nonNull(personDTO.getDisplayName()) && !personDTO.getDisplayName().equals(person.getDisplayName())) {
            person.setDisplayName(personDTO.getDisplayName());
        }
        if (nonNull(personDTO.getFirstName()) && !personDTO.getDisplayName().equals(person.getFirstName())) {
            person.setFirstName(personDTO.getFirstName());
        }
        if (nonNull(personDTO.getLastName()) && !personDTO.getLastName().equals(person.getLastName())) {
            person.setLastName(personDTO.getLastName());
        }
        personRepository.save(person);
    }

    public void deletePerson(Long personId) {
        var personPolicies = policyRepository.findAllByInsuredPersonId(personId);
        StreamSupport.stream(personPolicies.spliterator(), false)
                .forEach(policy -> {
                    policy.setInsuredPerson(null);
                    policyRepository.save(policy);
                });
        personRepository.deleteById(personId);
    }

}
