package pl.edu.amu.entity;

import jakarta.persistence.*;
import lombok.*;
import lombok.experimental.FieldDefaults;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.Set;

import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

@NoArgsConstructor
@AllArgsConstructor
@Builder(builderClassName = "EventBuilder", toBuilder = true)
@EqualsAndHashCode
@Getter
@Setter
@ToString
@FieldDefaults(level = AccessLevel.PUBLIC)
@Entity
public class Event implements Comparable<Event> {

    @Id
    @GeneratedValue(strategy = GenerationType.SEQUENCE)
    @Column(name = "event_id")
    Long eventId;

    @OneToMany(mappedBy = "event", fetch = FetchType.EAGER)
    Set<EventPerson> people;

    @Column(name = "event_year")
    Integer eventYear;

    @Column(name = "event_month")
    Integer eventMonth;

    @Column(name = "event_day")
    Integer eventDay;

    @Column(name = "event_calendar")
    @Enumerated(EnumType.ORDINAL)
    Calendar calendar;

    @ManyToOne
    @JoinColumn(name = "place_id")
    Place place;

    @Column(nullable = false)
    @Enumerated(EnumType.STRING)
    EventType type;

    @Column
    String description;

    @Column(columnDefinition = "boolean default false")
    Boolean isRecurring;

    @Column(name = "creation_date", nullable = false)
    ZonedDateTime creationDate;

    public boolean hasFullDate() {
        return !(isNull(eventDay) || isNull(eventMonth) || isNull(eventYear));
    }

    @Override
    public int compareTo(Event other) {
        if (isNull(other)) {
            return -1;
        }
        if (hasFullDate() && other.hasFullDate()) {
            var date = LocalDate.of(eventYear, eventMonth, eventDay);
            var otherDate = LocalDate.of(other.eventYear, other.eventMonth, other.eventDay);
            return date.compareTo(otherDate);
        }
        if (nonNull(eventYear) && nonNull(other.eventYear)) {
            return Integer.compare(eventYear, other.eventYear);
        }
        if (isNull(eventYear) && isNull(other.eventYear)) {
            return 0;
        }
        if (isNull(eventYear)) {
            return 1;
        }
        return -1;
    }

    public static final class EventBuilder {
        Boolean isRecurring = false;
        ZonedDateTime creationDate = ZonedDateTime.now();
    }

}
