package pl.edu.amu.service;

import com.google.gson.GsonBuilder;
import jakarta.transaction.Transactional;
import lombok.AccessLevel;
import lombok.experimental.FieldDefaults;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import pl.edu.amu.entity.*;
import pl.edu.amu.repository.*;
import pl.edu.amu.util.ZonedDateTimeGsonSerializer;
import pl.edu.amu.web.data.AncestorsData;
import pl.edu.amu.web.data.EventUpsert;
import pl.edu.amu.web.data.PersonDisplayName;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Objects.requireNonNull;
import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toMap;

@Service
@FieldDefaults(makeFinal = true, level = AccessLevel.PRIVATE)
@Slf4j
@Transactional
public class DbService {

    AncestorRepository ancestorRepository;
    PersonRepository personRepository;
    EventRepository eventRepository;
    FamilyMemberRepository familyMemberRepository;
    EventPersonRepository eventPersonRepository;

    @Autowired
    public DbService(
            AncestorRepository ancestorRepository,
            PersonRepository personRepository,
            EventPersonRepository eventPersonRepository,
            EventRepository eventRepository,
            FamilyMemberRepository familyMemberRepository
    ) {
        this.ancestorRepository = ancestorRepository;
        this.personRepository = personRepository;
        this.eventPersonRepository = eventPersonRepository;
        this.eventRepository = eventRepository;
        this.familyMemberRepository = familyMemberRepository;
    }

    public List<AncestorsData> findAncestorsData(Long personId) {
        var ancestorForPerson = ancestorRepository.findAllByPersonId(personId);
        var ancestorToFamilyId = findAncestorFamilies(personId);
        var ancestorsByFamilies = ancestorForPerson.stream()
                .collect(groupingBy(ancestor -> ancestorToFamilyId.getOrDefault(ancestor.getAncestor().personId, -1L)));
        return ancestorsByFamilies.entrySet().stream().flatMap(this::toAncestorData).sorted().toList();
    }

    private Map<Long, Long> findAncestorFamilies(Long personId) {
        return familyMemberRepository.findAncestorFamily(personId).stream()
                .flatMap(family -> Stream.of(
                        Pair.of(family.getMotherId(), family.getFamilyId()),
                        Pair.of(family.getFatherId(), family.getFamilyId())
                )).collect(toMap(Pair::getLeft, Pair::getRight));
    }

    private Stream<AncestorsData> toAncestorData(Entry<Long, List<Ancestor>> family) {
        if (family.getKey() < 0) {
            return family.getValue().stream().map(toAncestorsData(Optional.empty()));
        }
        if (family.getValue().size() != 2) {
            throw new IllegalStateException(String.format("Invalid family size for family id: %d", family.getKey()));
        }
        var bride = family.getValue().get(0);
        var groom = family.getValue().get(1);
        var maybeMarriage = eventRepository.findMarriageByBrideGroom(bride.ancestor, groom.ancestor);
        return Stream.of(toAncestorData(bride, maybeMarriage), toAncestorData(groom, maybeMarriage));
    }

    private Function<Ancestor, AncestorsData> toAncestorsData(Optional<Event> maybeMarriage) {
        return ancestor -> toAncestorData(ancestor, maybeMarriage);
    }

    private AncestorsData toAncestorData(Ancestor ancestor, Optional<Event> maybeMarriage) {
        var events = eventRepository.findByPerson(ancestor.ancestor);
        return AncestorsData.builder()
                .personId(ancestor.ancestor.personId)
                .displayName(ancestor.ancestor.displayName)
                .sex(ancestor.ancestor.sex.name())
                .birth(extractSingleEvent(findEvent(events, EventType.BIRTH)))
                .marriage(maybeMarriage.orElse(null))
                .death(extractSingleEvent(findEvent(events, EventType.DEATH)))
                .level(ancestor.level)
                .ancestorOffset(ancestor.ancestorOffset)
                .build();
    }

    private List<Event> findEvent(List<Event> events, EventType type) {
        return events.stream()
                .filter(ev -> ev.getType() == type)
                .toList();
    }

    private Event extractSingleEvent(List<Event> events) {
        return events.isEmpty() ? null : events.get(0);
    }


    public void upsertEvent(EventUpsert data, Long descendantId) {
        if (data.eventType == EventType.BIRTH || data.eventType == EventType.DEATH) {
            upsertSingularEvent(data);
            return;
        }
        if (data.eventType != EventType.MARRIAGE) {
            throw new IllegalStateException(String.format("Don't know how to update %s event.", data.eventType));
        }
        upsertMarriage(data, descendantId);
    }

    private void upsertSingularEvent(EventUpsert data) {
        var person = personRepository.findById(data.personId).orElseThrow();
        var maybeEvent = eventRepository.findByPerson(person).stream()
                .filter(ev -> ev.type == data.eventType)
                .findAny();
        if (maybeEvent.isPresent()) {
            updateSingularEvent(maybeEvent.get(), data);
            return;
        }
        insertEvent(person, data);
    }

    private void updateSingularEvent(Event event, EventUpsert data) {
        event.eventYear = data.eventDate.getYear();
        event.eventMonth = data.eventDate.getMonthValue();
        event.eventDay = data.eventDate.getDayOfMonth();
        eventRepository.save(event);
    }

    private void insertEvent(Person person, EventUpsert data) {
        var event = eventRepository.save(Event.builder()
                .eventYear(data.eventDate.getYear())
                .eventMonth(data.eventDate.getMonthValue())
                .eventDay(data.eventDate.getDayOfMonth())
                .calendar(pl.edu.amu.entity.Calendar.GREGORIAN)
                .type(data.eventType)
                .build());
        eventPersonRepository.insert(
                event.eventId, person.personId, EventPersonRole.SUBJECT.name(), null, null, null, null
        );
    }

    private void upsertMarriage(EventUpsert data, Long descendantId) {
        var ancestorFamilies = familyMemberRepository.findAncestorFamily(descendantId);
        var family = ancestorFamilies.stream().filter(byPerson(data.personId)).findFirst().orElseThrow();
        var mother = personRepository.findById(family.getMotherId()).orElseThrow();
        var father = personRepository.findById(family.getFatherId()).orElseThrow();
        var maybeMarriage = eventRepository.findMarriageByBrideGroom(mother, father);
        if (maybeMarriage.isPresent()) {
            log.info("Found marriage: {}, updating.", maybeMarriage.get().eventId);
            updateSingularEvent(maybeMarriage.get(), data);
            return;
        }
        log.info("Marriage not found, inserting marriage between {} and {}.", mother.displayName, father.displayName);
        var event = eventRepository.save(Event.builder()
                .eventYear(data.eventDate.getYear())
                .eventMonth(data.eventDate.getMonthValue())
                .eventDay(data.eventDate.getDayOfMonth())
                .calendar(Calendar.GREGORIAN)
                .type(data.eventType)
                .build());
        eventPersonRepository.insert(
                event.eventId, mother.personId, EventPersonRole.BRIDE.name(), null, null, null, null
        );
        eventPersonRepository.insert(
                event.eventId, father.personId, EventPersonRole.GROOM.name(), null, null, null, null
        );
    }

    private Predicate<AncestorFamily> byPerson(Long personId) {
        return family -> personId.equals(family.getMotherId()) || personId.equals(family.getMotherId());
    }


    public void updateDisplayName(PersonDisplayName personDisplayName) {
        var person = personRepository.findById(personDisplayName.personId).orElseThrow();
        person.displayName = requireNonNull(personDisplayName.displayName);
    }

    public byte[] getAllData() {
        // var gson = new GsonBuilder().registerTypeAdapter(ZonedDateTime.class, new ZonedDateTimeGsonSerializer()).create();
        // We need custom type adapters here, as there is a cycle in the data, causing StackOverflow.
        // Too late for me to cover it, sorry.
        var ancestors = """
                { "name": "Test dump, does not work :("}
                """;
        //gson.toJson(ancestorRepository.findAll());
        return ancestors.getBytes(UTF_8);
    }

}
